/*----- PROTECTED REGION ID(TestDevice.cpp) ENABLED START -----*/

//=============================================================================
//
// file :        TestDevice.cpp
//
// description : C++ source for the TestDevice and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               TestDevice are implemented in this file.
//
// $Author: graziano $
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <TestDevice.h>
#include <TestDeviceClass.h>
#include <math.h>
#include <iomanip>
#include <limits>

/*----- PROTECTED REGION END -----*/


/**
 *	TestDevice class description:
 *	
 */

//================================================================
//
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name  |  Method name
//----------------------------------------------------------------
//  State         |  Inherited (no method)
//  Status        |  Inherited (no method)
//================================================================

namespace TestDevice_ns
{
	/*----- PROTECTED REGION ID(TestDevice::namespace_starting) ENABLED START -----*/

	//	static initializations

	/*----- PROTECTED REGION END -----*/	//	TestDevice::namespace_starting



//--------------------------------------------------------
/**
 *	Method      : TestDevice::TestDevice()
 *	Description : Constructors for a Tango device
 *	              implementing the class TestDevice
 */
//--------------------------------------------------------
TestDevice::TestDevice(Tango::DeviceClass *cl, string &s)
 	: TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(TestDevice::constructor_1) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	TestDevice::constructor_1
}
//--------------------------------------------------------
TestDevice::TestDevice(Tango::DeviceClass *cl, const char *s)
 	: TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(TestDevice::constructor_2) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	TestDevice::constructor_2
}
//--------------------------------------------------------
TestDevice::TestDevice(Tango::DeviceClass *cl, const char *s, const char *d)
 	: TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(TestDevice::constructor_3) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	TestDevice::constructor_3
}


//--------------------------------------------------------
/**
 *	Method      : TestDevice::delete_device()()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void TestDevice::delete_device()
{
	/*----- PROTECTED REGION ID(TestDevice::delete_device) ENABLED START -----*/

	//	Delete device allocated objects

	save_configuration();

	map<std::string,map_data_t>::iterator it;
	for(it = attr_data_map.begin(); it != attr_data_map.end(); it++)
	{
		if(it->second.type == Tango::DEV_DOUBLE)
			delete [] (Tango::DevDouble *) it->second.value;
		else if(it->second.type == Tango::DEV_LONG)
			delete [] (Tango::DevLong *) it->second.value;
		else if(it->second.type == Tango::DEV_BOOLEAN)
			delete [] (Tango::DevBoolean *) it->second.value;
		else if(it->second.type == Tango::DEV_STRING)
			delete [] (char *) it->second.value;	//TODO: !!!
	}

	/*----- PROTECTED REGION END -----*/	//	TestDevice::delete_device
	delete[] attr_ScalarDyn_read;
	delete[] attr_Ref_read;
	delete[] attr_Values_read;

}


//--------------------------------------------------------
/**
 *	Method      : TestDevice::init_device()
 *	Description : //	will be called at device initialization.
 */
//--------------------------------------------------------
void TestDevice::init_device()
{
	DEBUG_STREAM << "TestDevice::init_device() create device " << device_name << endl;

	/*----- PROTECTED REGION ID(TestDevice::init_device_before) ENABLED START -----*/

	//	Initialization before get_device_property() call
	dlastDB = 0.0;
	waitingDB = false;

	/*----- PROTECTED REGION END -----*/	//	TestDevice::init_device_before
	
	//	Get the device properties (if any) from database
	get_device_property();
	
	attr_ScalarDyn_read = new Tango::DevString[1];
	attr_Ref_read = new Tango::DevDouble[1024];
	attr_Values_read = new Tango::DevString[1024];
	
	/*----- PROTECTED REGION ID(TestDevice::init_device) ENABLED START -----*/
	for (int i = 0; i < 1024; i++)
	{
		//ref_ptr[i] = ref_array[i];
		attr_Values_read[i] = ref_array[i];
	}

	//	Initialize device
	gettimeofday(&last_read, NULL);
	set_state(Tango::ON);

	/*----- PROTECTED REGION END -----*/	//	TestDevice::init_device
}



//--------------------------------------------------------
/**
 *	Method      : TestDevice::get_device_property()
 *	Description : //	Add your own code to initialize
 */
//--------------------------------------------------------
void TestDevice::get_device_property()
{
	/*----- PROTECTED REGION ID(TestDevice::get_device_property_before) ENABLED START -----*/

	//	Initialize property data members

	/*----- PROTECTED REGION END -----*/	//	TestDevice::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("Attr_config"));
	dev_prop.push_back(Tango::DbDatum("MinSaveDBPeriod"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on TestDeviceClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		TestDeviceClass	*ds_class =
			(static_cast<TestDeviceClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize Attr_config from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  attr_config;
		else {
			//	Try to initialize Attr_config from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  attr_config;
		}
		//	And try to extract Attr_config value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  attr_config;

		//	Try to initialize MinSaveDBPeriod from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  minSaveDBPeriod;
		else {
			//	Try to initialize MinSaveDBPeriod from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  minSaveDBPeriod;
		}
		//	And try to extract MinSaveDBPeriod value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  minSaveDBPeriod;


	}
	/*----- PROTECTED REGION ID(TestDevice::get_device_property_after) ENABLED START -----*/

	//	Check device property data members init

	/*----- PROTECTED REGION END -----*/	//	TestDevice::get_device_property_after

}

//--------------------------------------------------------
/**
 *	Method      : TestDevice::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void TestDevice::always_executed_hook()
{
	//INFO_STREAM << "TestDevice::always_executed_hook()  " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::always_executed_hook) ENABLED START -----*/

	//	code always executed before all requests

	/*----- PROTECTED REGION END -----*/	//	TestDevice::always_executed_hook
}



//--------------------------------------------------------
/**
 *	Method      : TestDevice::read_attr_hardware()
 *	Description : Hardware acquisition for attributes.
 */
//--------------------------------------------------------
void TestDevice::read_attr_hardware(vector<long> &attr_list)
{
	//DEBUG_STREAM << "TestDevice::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(TestDevice::read_attr_hardware) ENABLED START -----*/

	//	Add your own code
	if(waitingDB)
	{
		timeval now;
		gettimeofday(&now, NULL);
		double dnow=now.tv_sec + (double)now.tv_usec/1000000.0;
		if(dnow - dlastDB > minSaveDBPeriod)
		{
			save_configuration();
			dlastDB = dnow;
			waitingDB = false;
		}
	}

	/*----- PROTECTED REGION END -----*/	//	TestDevice::read_attr_hardware

}


//--------------------------------------------------------
/**
 *	Read ScalarDyn attribute
 *	Description:
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TestDevice::read_ScalarDyn(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "TestDevice::read_ScalarDyn(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TestDevice::read_ScalarDyn) ENABLED START -----*/

	map<std::string,map_data_t>::iterator it = attr_data_map.find(attr.get_name());
	if(it == attr_data_map.end())
		Tango::Except::throw_exception((const char*) "Exception reading attribute",
				(const char*) "Attribute Not Found", __FUNCTION__, Tango::ERR);

	if(it->second.type == Tango::DEV_DOUBLE)
	{
		attr.set_value_date_quality((Tango::DevDouble *) it->second.value,it->second.ts,Tango::ATTR_VALID, 1, 0, false);
	}
	else if(it->second.type == Tango::DEV_LONG)
	{
		attr.set_value_date_quality((Tango::DevLong *) it->second.value,it->second.ts,Tango::ATTR_VALID, 1, 0, false);
	}
	else if(it->second.type == Tango::DEV_BOOLEAN)
	{
		attr.set_value_date_quality((Tango::DevBoolean *) it->second.value,it->second.ts,Tango::ATTR_VALID, 1, 0, false);
	}
	else if(it->second.type == Tango::DEV_STRING)
	{
		*attr_ScalarDyn_read = (char *) it->second.value;
		attr.set_value_date_quality(attr_ScalarDyn_read,it->second.ts,Tango::ATTR_VALID, 1, 0, false);
		//attr.set_value_date_quality((Tango::DevString *) it->second.value,it->second.ts,Tango::ATTR_VALID, 1, 0, false);
	}

	/*----- PROTECTED REGION END -----*/	//	TestDevice::read_ScalarDyn
}
//+----------------------------------------------------------------------------
//
// method : 		TestDevice::write_ScalarDyn
//
// description : 	Write ScalarDyn attribute values to hardware.
//
//-----------------------------------------------------------------------------
void TestDevice::write_ScalarDyn(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "TestDevice::write_ScalarDyn(Tango::WAttribute &attr) entering... "<< endl;
	/*----- PROTECTED REGION ID(TestDevice::write_ScalarDyn) ENABLED START -----*/

	map<std::string,map_data_t>::iterator it = attr_data_map.find(attr.get_name());
	if(it == attr_data_map.end())
		Tango::Except::throw_exception((const char*) "Exception reading attribute",
				(const char*) "Attribute Not Found", __FUNCTION__, Tango::ERR);

	gettimeofday(&(it->second.ts), NULL);
	if(it->second.type == Tango::DEV_DOUBLE)
	{
		Tango::DevDouble w_val;
		attr.get_write_value(w_val);
		*(Tango::DevDouble *) it->second.value = w_val;
		try{
			push_change_event(attr.get_name(),(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
			push_archive_event(attr.get_name(),(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
		}catch(...){}
	}
	else if(it->second.type == Tango::DEV_LONG)
	{
		Tango::DevLong w_val;
		attr.get_write_value(w_val);
		*(Tango::DevLong *) it->second.value = w_val;
		try{
			push_change_event(attr.get_name(),(Tango::DevLong *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
			push_archive_event(attr.get_name(),(Tango::DevLong *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
		}catch(...){}
	}
	else if(it->second.type == Tango::DEV_BOOLEAN)
	{
		Tango::DevBoolean w_val;
		attr.get_write_value(w_val);
		*(Tango::DevBoolean *) it->second.value = w_val;
		try{
			push_change_event(attr.get_name(),(Tango::DevBoolean *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
			push_archive_event(attr.get_name(),(Tango::DevBoolean *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
		}catch(...){}
	}
	else if(it->second.type == Tango::DEV_STRING)
	{
		Tango::DevString w_val;
		attr.get_write_value(w_val);
		memset((char *) it->second.value, 0, 1024);
		strncpy((char *) it->second.value,w_val, 1023);
		//try{
		//	push_change_event(attr.get_name(),(Tango::DevString *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
		//	push_archive_event(attr.get_name(),(Tango::DevString *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
		//}catch(...){}
	}
	timeval now;
	gettimeofday(&now, NULL);
	double dnow=now.tv_sec + (double)now.tv_usec/1000000.0;
	if(dnow - dlastDB > minSaveDBPeriod)
	{
		save_configuration();
		dlastDB = dnow;
		waitingDB = false;
	}
	else
	{
		waitingDB = true;
	}

	/*----- PROTECTED REGION END -----*/	//	TestDevice::write_ScalarDyn
}

//--------------------------------------------------------
/**
 *	Read SpectrumDyn attribute
 *	Description:
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum
 */
//--------------------------------------------------------
void TestDevice::read_SpectrumDyn(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "TestDevice::read_SpectrumDyn(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TestDevice::read_SpectrumDyn) ENABLED START -----*/

	map<std::string,map_data_t>::iterator it = attr_data_map.find(attr.get_name());
	if(it == attr_data_map.end())
		Tango::Except::throw_exception((const char*) "Exception reading attribute",
				(const char*) "Attribute Not Found", __FUNCTION__, Tango::ERR);

	if(it->second.type == Tango::DEV_DOUBLE)
	{
		attr.set_value_date_quality((Tango::DevDouble *) it->second.value,it->second.ts,Tango::ATTR_VALID, it->second.size1, 0, false);
	}
	else if(it->second.type == Tango::DEV_LONG)
	{
		attr.set_value_date_quality((Tango::DevLong *) it->second.value,it->second.ts,Tango::ATTR_VALID, it->second.size1, 0, false);
	}
	else if(it->second.type == Tango::DEV_BOOLEAN)
	{
		attr.set_value_date_quality((Tango::DevBoolean *) it->second.value,it->second.ts,Tango::ATTR_VALID, it->second.size1, 0, false);
	}
	else if(it->second.type == Tango::DEV_STRING)
	{
		//NOT SUPPORTED
	}

	/*----- PROTECTED REGION END -----*/	//	TestDevice::read_SpectrumDyn
}
//+----------------------------------------------------------------------------
//
// method : 		TestDevice::write_SpectrumDyn
//
// description : 	Write SpectrumDyn attribute values to hardware.
//
//-----------------------------------------------------------------------------
void TestDevice::write_SpectrumDyn(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "TestDevice::write_SpectrumDyn(Tango::WAttribute &attr) entering... "<< endl;
	/*----- PROTECTED REGION ID(TestDevice::write_SpectrumDyn) ENABLED START -----*/

	map<std::string,map_data_t>::iterator it = attr_data_map.find(attr.get_name());
	if(it == attr_data_map.end())
		Tango::Except::throw_exception((const char*) "Exception reading attribute",
				(const char*) "Attribute Not Found", __FUNCTION__, Tango::ERR);

	gettimeofday(&(it->second.ts), NULL);
	size_t value_size = attr.get_write_value_length();
	it->second.size1 = value_size > MAX_SPECTRUM_SIZE ? MAX_SPECTRUM_SIZE : value_size;
	if(it->second.type == Tango::DEV_DOUBLE)
	{
		const Tango::DevDouble * w_val;
		attr.get_write_value(w_val);
		for(size_t ind=0; ind < it->second.size1; ind++)
		{
			((Tango::DevDouble *) it->second.value)[ind] = w_val[ind];
		}
		try{
			push_change_event(attr.get_name(),(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1, 0, false);
			push_archive_event(attr.get_name(),(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1, 0, false);
		}catch(...){}
	}
	else if(it->second.type == Tango::DEV_LONG)
	{
		const Tango::DevLong * w_val;
		attr.get_write_value(w_val);
		for(size_t ind=0; ind < it->second.size1; ind++)
		{
			((Tango::DevLong *) it->second.value)[ind] = w_val[ind];
		}
		try{
			push_change_event(attr.get_name(),(Tango::DevLong *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1, 0, false);
			push_archive_event(attr.get_name(),(Tango::DevLong *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1, 0, false);
		}catch(...){}
	}
	else if(it->second.type == Tango::DEV_BOOLEAN)
	{
		const Tango::DevBoolean * w_val;
		attr.get_write_value(w_val);
		for(size_t ind=0; ind<it->second.size1; ind++)
		{
			((Tango::DevBoolean *) it->second.value)[ind] = w_val[ind];
		}
		try{
			push_change_event(attr.get_name(),(Tango::DevBoolean *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1, 0, false);
			push_archive_event(attr.get_name(),(Tango::DevBoolean *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1, 0, false);
		}catch(...){}
	}
	else if(it->second.type == Tango::DEV_STRING)
	{
		//NOT SUPPORTED
	}
	timeval now;
	gettimeofday(&now, NULL);
	double dnow=now.tv_sec + (double)now.tv_usec/1000000.0;
	if(dnow - dlastDB > minSaveDBPeriod)
	{
		save_configuration();
		dlastDB = dnow;
		waitingDB = false;
	}
	else
	{
		waitingDB = true;
	}

	/*----- PROTECTED REGION END -----*/	//	TestDevice::write_SpectrumDyn
}

//--------------------------------------------------------
/**
 *	Read Ref attribute
 *	Description:
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TestDevice::read_Ref(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "TestDevice::read_Ref(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TestDevice::read_Ref) ENABLED START -----*/

	map<std::string,map_data_t>::iterator it;
	int i=0;
	for(it = attr_data_map.begin(), i=0; it != attr_data_map.end() && i<1024; it++)
	{
		if(it->second.type == Tango::DEV_DOUBLE)
			attr_Ref_read[i++] =  *((Tango::DevDouble *) it->second.value);
	}
	attr.set_value(attr_Ref_read, i);
	/*----- PROTECTED REGION END -----*/	//	TestDevice::read_Ref
}
//+----------------------------------------------------------------------------
//
// method : 		TestDevice::write_Ref
//
// description : 	Write Ref attribute values to hardware.
//
//-----------------------------------------------------------------------------
void TestDevice::write_Ref(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "TestDevice::write_Ref(Tango::Attribute &attr) entering... " << endl;

	//	Retrieve write value
	const Tango::DevDouble *w_val;
	attr.get_write_value(w_val);

	/*----- PROTECTED REGION ID(TestDevice::write_Ref) ENABLED START -----*/
	int size_ref = attr.get_write_value_length();
	size_ref =  size_ref<1024 ? size_ref : 1024;
	double tmp;
	int i=0;
	attr.get_write_value(tmp);

	for(map<std::string,map_data_t>::iterator it=attr_data_map.begin(); it != attr_data_map.end() && i<size_ref; it++)
	{
		if(it->second.type == Tango::DEV_DOUBLE)
		{
			gettimeofday(&(it->second.ts), NULL);
			*(Tango::DevDouble *) it->second.value = w_val[i++];
			try {
				push_change_event(it->first,(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
				push_archive_event(it->first,(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
			}catch(...){}
		}
	}
	timeval now;
	gettimeofday(&now, NULL);
	double dnow=now.tv_sec + (double)now.tv_usec/1000000.0;
	if(dnow - dlastDB > minSaveDBPeriod)
	{
		save_configuration();
		dlastDB = dnow;
		waitingDB = false;
	}
	else
	{
		waitingDB = true;
	}

	/*----- PROTECTED REGION END -----*///	TestDevice::write_Ref
}

//--------------------------------------------------------
/**
 *	Read Values attribute
 *	Description:
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void TestDevice::read_Values(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "TestDevice::read_Values(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(TestDevice::read_Values) ENABLED START -----*/

	map<std::string,map_data_t>::iterator it;
	int i;
	for(it = attr_data_map.begin(), i=0; it != attr_data_map.end() && i<1024; it++, i++)
	{
		stringstream val;
		if(it->second.type == Tango::DEV_DOUBLE)
		{
			for(size_t ind=0; ind<it->second.size1; ind++)
			{
				val << std::setprecision(std::numeric_limits<Tango::DevDouble>::digits10+2) << ((Tango::DevDouble *) it->second.value)[ind];
				if(ind < it->second.size1-1)
					val << ",";
			}
			sprintf(ref_array[i], "%s=%s", it->first.c_str(), val.str().c_str());
		}
		else if(it->second.type == Tango::DEV_LONG)
		{
			for(size_t ind=0; ind<it->second.size1; ind++)
			{
				val << std::setprecision(std::numeric_limits<Tango::DevLong>::digits10+2) << ((Tango::DevLong *) it->second.value)[ind];
				if(ind < it->second.size1-1)
					val << ",";
			}
			sprintf(ref_array[i], "%s=%s", it->first.c_str(), val.str().c_str());
		}
		else if(it->second.type == Tango::DEV_BOOLEAN)
		{
			for(size_t ind=0; ind<it->second.size1; ind++)
			{
				val << (((Tango::DevBoolean *) it->second.value)[ind] ? "true" : "false");
				if(ind < it->second.size1-1)
					val << ",";
			}
			sprintf(ref_array[i], "%s=%s", it->first.c_str(), val.str().c_str());
		}
		else if(it->second.type == Tango::DEV_STRING)
		{
			sprintf(ref_array[i], "%s=%s", it->first.c_str(), ((char *) it->second.value));
		}

	}
	attr.set_value(attr_Values_read, i);
	/*----- PROTECTED REGION END -----*/	//	TestDevice::read_Values
}
//+----------------------------------------------------------------------------
//
// method : 		TestDevice::write_Ref
//
// description : 	Write Ref attribute values to hardware.
//
//-----------------------------------------------------------------------------
void TestDevice::write_Values(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "TestDevice::write_Values(Tango::Attribute &attr) entering... " << endl;

	//	Retrieve write value
	const Tango::ConstDevString *w_val;
	attr.get_write_value(w_val);

	/*----- PROTECTED REGION ID(TestDevice::write_Values) ENABLED START -----*/
	int size_ref = attr.get_write_value_length();
	size_ref =  size_ref<1024 ? size_ref : 1024;
	double tmp;
	int i=0;
	attr.get_write_value(tmp);

	for(i=0; i<size_ref; i++)
	{
		vector<string> results;
		string_explode(w_val[i], "=", &results);
		if(results.size() == 2)
		{
			map<std::string,map_data_t>::iterator it = attr_data_map.find(results[0]);

			if(it == attr_data_map.end())
				continue;
				/*Tango::Except::throw_exception((const char*) "Exception writing attribute",
						string("Attribute '") + string(results[0]) + string("' not found"), __FUNCTION__, Tango::ERR);*/
			gettimeofday(&(it->second.ts), NULL);
			vector<string> values;
			string_explode(results[1], ",", &values);
			if(it->second.type == Tango::DEV_BOOLEAN)
			{
				size_t ind=0;
				for(vector<string>::iterator val=values.begin(); val != values.end() && ind<it->second.size1; val++)
				{
					if(*val == "true")
						((Tango::DevBoolean *) it->second.value)[ind++] = true;
					else if(*val == "false")
						((Tango::DevBoolean *) it->second.value)[ind++] = false;
					else
						((Tango::DevBoolean *) it->second.value)[ind++] = atoi(val->c_str());
				}
				try {
					set_write_value((Tango::DevBoolean *) it->second.value, results[0], it->second.size1);
					push_change_event(it->first,(Tango::DevBoolean *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1/*size*/, 0, false);
					push_archive_event(it->first,(Tango::DevBoolean *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ it->second.size1/*size*/, 0, false);
				}catch(...){}
			}
			else if(it->second.type == Tango::DEV_LONG)
			{
				size_t ind=0;
				for(vector<string>::iterator val=values.begin(); val != values.end() && ind<it->second.size1; val++)
				{
					((Tango::DevLong *) it->second.value)[ind++] = atoi(results[1].c_str());
				}
				try {
					set_write_value((Tango::DevLong *) it->second.value, results[0], it->second.size1);
					push_change_event(it->first,(Tango::DevLong *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
					push_archive_event(it->first,(Tango::DevLong *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
				}catch(...){}
			}
			else if(it->second.type == Tango::DEV_DOUBLE)
			{
				size_t ind=0;
				for(vector<string>::iterator val=values.begin(); val != values.end() && ind<it->second.size1; val++)
				{
					((Tango::DevDouble *) it->second.value)[ind++] = atof(results[1].c_str());
				}
				try {
					set_write_value((Tango::DevDouble *) it->second.value, results[0], it->second.size1);
					push_change_event(it->first,(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
					push_archive_event(it->first,(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
				}catch(...){}
			}
			else if(it->second.type == Tango::DEV_STRING)
			{
				memset((char *) it->second.value, 0, 1024);
				strncpy((char *) it->second.value,results[1].c_str(), 1023);
				Tango::DevString str_ptr = (char *) it->second.value;
				set_write_value((Tango::DevString ) str_ptr, results[0]);
			}
		}
	}
	timeval now;
	gettimeofday(&now, NULL);
	double dnow=now.tv_sec + (double)now.tv_usec/1000000.0;
	if(dnow - dlastDB > minSaveDBPeriod)
	{
		save_configuration();
		dlastDB = dnow;
		waitingDB = false;
	}
	else
	{
		waitingDB = true;
	}
	//and now push events

	/*----- PROTECTED REGION END -----*///	TestDevice::write_Ref
}

//--------------------------------------------------------
/**
 *	Method      : TestDevice::TestDeviceClass::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *	              for specified device.
 */
//--------------------------------------------------------
void TestDevice::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(TestDevice::Class::add_dynamic_attributes) ENABLED START -----*/

	//	Add your own code to create and add dynamic attributes if any
	DEBUG_STREAM << __FUNCTION__ << " entering...";
    if (attr_config.empty() == false)
    {
		for(vector<string>::iterator it = attr_config.begin();it != attr_config.end(); it++)
		{
			vector<string> results;
			string_explode(*it, ":", &results);
			try
			{
				create_dynamic_attribute(results);
			} catch(Tango::DevFailed &e)
			{
				INFO_STREAM << __FUNCTION__ << " Error reading creating attribute='" << e.errors[0].desc << "'";
			}
		}
    }

	/*----- PROTECTED REGION END -----*/	//	TestDevice::Class::add_dynamic_attributes

}



//========================================================
//	Command execution methods
//========================================================

//--------------------------------------------------------
/**
 *	Execute the WriteAttrname command:
 *	Description: Delete all Double attributes and create new ones.
 *
 *	@param argin List of double attributes to write.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::write_attrname(const Tango::DevVarDoubleStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::write_attrname()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::write_attrname) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here

	double tmp;
	int lens, lend, lenmin;
	string name;
	int i=0;
	lens = argin->svalue.length();
	lend = argin->dvalue.length();
	lenmin = (lens <= lend) ? lens : lend;

	for(i=0; i<lenmin; i++)
	{
		name = string(argin->svalue[i]);
		tmp = argin->dvalue[i];

		map<std::string,map_data_t>::iterator it = attr_data_map.find(name);
		if(it == attr_data_map.end())
			Tango::Except::throw_exception((const char*) "Exception writing attribute", string("Attribute '") + name
								+ string("' not found"), __FUNCTION__, Tango::ERR);

		if(it->second.type != Tango::DEV_DOUBLE)
		{
			Tango::Except::throw_exception((const char*) "Exception writing attribute", string("Attribute '") + name
											+ string("' is not of type double"), __FUNCTION__, Tango::ERR);
		}
		else
		{
			gettimeofday(&(it->second.ts), NULL);
			*(Tango::DevDouble *) it->second.value = tmp;
			try {
				push_change_event(it->first,(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
				push_archive_event(it->first,(Tango::DevDouble *)(it->second.value),/*NULL,Tango::ATTR_VALID,*/ 1/*size*/, 0, false);
			}catch(...){}
			timeval now;
			gettimeofday(&now, NULL);
			double dnow=now.tv_sec + (double)now.tv_usec/1000000.0;
			if(dnow - dlastDB > minSaveDBPeriod)
			{
				save_configuration();
				dlastDB = dnow;
				waitingDB = false;
			}
			else
			{
				waitingDB = true;
			}
		}
	}

	/*----- PROTECTED REGION END -----*///	TestDevice::write_attrname

	return;
}

//--------------------------------------------------------
/**
 *	Execute the ReadAttrname command:
 *	Description: Delete all Double attributes and create new ones.
 *
 *	@param argin List of new double attributes.
 *	@returns
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray * TestDevice::read_attrname(const Tango::DevVarStringArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "TestDevice::read_attrname()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::read_attrname) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	argout = new Tango::DevVarDoubleArray();
	vector<string> attributes;
	attributes << (*argin);
	double *tmp = new double[attributes.size()];
	string name;

	unsigned int index = 0;

	for(vector<string>::iterator its = attributes.begin(); its != attributes.end(); its++)
	{
		name = *its;
		map<std::string, map_data_t>::iterator it = attr_data_map.find(name);
		if(it == attr_data_map.end())
		{
			delete [] tmp;
			Tango::Except::throw_exception((const char*) "Exception reading attribute", string("Attribute '") + name
					+ string("' not found"), __FUNCTION__, Tango::ERR);
		}
		if(it->second.type == Tango::DEV_DOUBLE)
		{
			tmp[index++] = *((Tango::DevDouble *) it->second.value);
		}
		else
		{
			delete [] tmp;
			Tango::Except::throw_exception((const char*) "Exception reading attribute", string("Attribute '") + name
								+ string("' is not of type double"), __FUNCTION__, Tango::ERR);
		}
	}
	//if found all names
	if(index == attributes.size())
	{
		argout->length(index);
		for(unsigned int i=0; i<index; i++)
			(*argout)[i] = tmp[i];
	}
	delete [] tmp;
	/*----- PROTECTED REGION END -----*///	TestDevice::read_attrname

	return argout;
}

//--------------------------------------------------------
/**
 *	Execute the Config command:
 *	Description: Delete all Double attributes and create new ones.
 *
 *	@param argin List of new double attributes.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::config(const Tango::DevVarStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::config()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::config) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	vector<string> attributes;
	attributes << (*argin);

#if 1
	//TODO: verify memory leak
	for(unsigned int i = 0; i < this->get_device_attr()->get_attr_nb(); i++)
	{
		string attr_name = get_device_attr()->get_attr_by_ind(i).get_name();
		if((attr_name != "State") && (attr_name != "Status") && (attr_name != "Ref")  && (attr_name != "Values"))
		{
			DEBUG_STREAM << "TestDevice::config: i=" << i << " removing attribute: "
						<< attr_name << endl;
			//TODO: update_idx=Flag set to true if the attributes object index
			//		used to retrieve the corresponding Attr object has to be
			//		updated (because one Attr object has been removed)
			this->get_device_attr()->remove_attribute(attr_name, /*update_idx*/true);
			i--;
		}
	}
#else
	erase_dynamic_attributes()	//TODO
#endif

	map<std::string,map_data_t>::iterator it;
	for(it = attr_data_map.begin(); it != attr_data_map.end(); it++)
	{
		if(it->second.type == Tango::DEV_DOUBLE)
			delete [] (Tango::DevDouble *) it->second.value;
		else if(it->second.type == Tango::DEV_LONG)
			delete [] (Tango::DevLong *) it->second.value;
		else if(it->second.type == Tango::DEV_BOOLEAN)
			delete [] (Tango::DevBoolean *) it->second.value;
		else if(it->second.type == Tango::DEV_STRING)
			delete [] (char *) it->second.value;
	}

	try
	{
		attr_data_map.clear();
		DEBUG_STREAM << "TestDevice::config: attr_data_map cleared !" << endl;
	} catch(...)
	{
		WARN_STREAM << "TestDevice::delete_device: Could not clear attr_data_map !" << endl;
	}

	numAttr = attributes.size();
	for(int j = 0; j < numAttr; j++)
	{
		Tango::Attr  *cspec;
		char attrname[256];
		Tango::UserDefaultAttrProp attrprop;

		sprintf(attrname, "%s", attributes[j].c_str());
		cspec = new ScalarDynAttrib(attrname, Tango::DEV_DOUBLE);
		/*sprintf(strtmp,"Dynamic Attribute %s reading channel %d ",attrname,num_ch);
		 if(unit != "")
		 {
		 attrprop.set_unit(unit.c_str());
		 attrprop.set_standard_unit("1.0");
		 attrprop.set_display_unit(unit.c_str());
		 }
		 attrprop.set_format("%.9g");
		 attrprop.set_description(strtmp);*/
		cspec->set_default_properties(attrprop);
		add_attribute(cspec);
		struct timeval curr_time;
		//gettimeofday(&curr_time,NULL);
		curr_time.tv_sec = 0;

		double *value = new double();
		map_data_t tmp;
		tmp.value = value;
		tmp.size1 = 1;
		tmp.size2 = 0;
		tmp.max1 = 1;
		tmp.max2 = 1;
		tmp.type = Tango::DEV_DOUBLE;
		tmp.ts = curr_time;
		attr_data_map.insert(make_pair(attrname, tmp));
	}
	save_configuration();
	/*----- PROTECTED REGION END -----*///	TestDevice::config
	return;
}

//--------------------------------------------------------
/**
 *	Execute the Add command:
 *	Description: Create new attributes.
 *
 *	@param argin List of new attributes.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::add(const Tango::DevVarStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::add()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::add) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	vector<string> attributes;
	attributes << (*argin);

	for(vector<string>::iterator it = attributes.begin();it != attributes.end(); it++)
	{
		DEBUG_STREAM << __func__ << ": adding '" << *it << "'";
		vector<string> results;
		string_explode(*it, ":", &results);
		create_dynamic_attribute(results);
	}
	save_configuration();

	/*----- PROTECTED REGION END -----*///	TestDevice::add
	return;
}

//--------------------------------------------------------
/**
 *	Execute the AddDouble command:
 *	Description: Create new attributes.
 *
 *	@param argin List of new attributes.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::add_double(const Tango::DevVarStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::add_double()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::add_double) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	vector<string> attributes;
	attributes << (*argin);

	for(vector<string>::iterator it = attributes.begin();it != attributes.end(); it++)
	{
		vector<string> results;
		results.push_back(*it);
		results.push_back("double");
		results.push_back("0");
		results.push_back("0");
		create_dynamic_attribute(results);
	}
	save_configuration();

	/*----- PROTECTED REGION END -----*///	TestDevice::add_double
	return;
}

//--------------------------------------------------------
/**
 *	Execute the AddLong command:
 *	Description: Create new attributes.
 *
 *	@param argin List of new attributes.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::add_long(const Tango::DevVarStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::add_long()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::add_long) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	vector<string> attributes;
	attributes << (*argin);

	for(vector<string>::iterator it = attributes.begin();it != attributes.end(); it++)
	{
		vector<string> results;
		results.push_back(*it);
		results.push_back("long");
		results.push_back("0");
		results.push_back("0");
		create_dynamic_attribute(results);
	}
	save_configuration();

	/*----- PROTECTED REGION END -----*///	TestDevice::add_long
	return;
}

//--------------------------------------------------------
/**
 *	Execute the AddBool command:
 *	Description: Create new attributes.
 *
 *	@param argin List of new attributes.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::add_bool(const Tango::DevVarStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::add_bool()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::add_bool) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	vector<string> attributes;
	attributes << (*argin);

	for(vector<string>::iterator it = attributes.begin();it != attributes.end(); it++)
	{
		vector<string> results;
		results.push_back(*it);
		results.push_back("bool");
		results.push_back("0");
		results.push_back("0");
		create_dynamic_attribute(results);
	}
	save_configuration();

	/*----- PROTECTED REGION END -----*///	TestDevice::add_bool
	return;
}

//--------------------------------------------------------
/**
 *	Execute the AddString command:
 *	Description: Create new attributes.
 *
 *	@param argin List of new attributes.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::add_string(const Tango::DevVarStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::add_string()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::add_string) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	vector<string> attributes;
	attributes << (*argin);

	for(vector<string>::iterator it = attributes.begin();it != attributes.end(); it++)
	{
		vector<string> results;
		results.push_back(*it);
		results.push_back("string");
		results.push_back("0");
		results.push_back("0");
		create_dynamic_attribute(results);
	}
	save_configuration();

	/*----- PROTECTED REGION END -----*///	TestDevice::add_string
	return;
}

//--------------------------------------------------------
/**
 *	Execute the Remove command:
 *	Description: Remove attributes.
 *
 *	@param argin List of attributes.
 *	@returns
 */
//--------------------------------------------------------
void TestDevice::remove(const Tango::DevVarStringArray *argin)
{
	DEBUG_STREAM << "TestDevice::remove()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(TestDevice::remove) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

	//	Add your own code to control device here
	vector<string> attributes;
	attributes << (*argin);

	//TODO: verify memory leak
	for(vector<string>::iterator it = attributes.begin(); it != attributes.end(); it++)
	{

		if((*it != "State") && (*it != "Status") && (*it != "Ref"))
		{
			map<std::string,map_data_t>::iterator it_attr = attr_data_map.find(*it);
			if(it_attr == attr_data_map.end())
			{
				Tango::Except::throw_exception((const char*) "Exception removing attribute", string("Attribute '")
						+ string(*it) + string("' not found"), __FUNCTION__, Tango::ERR);
			}
			DEBUG_STREAM << __FUNCTION__ << ": removing attribute: " << *it;
			//TODO: update_idx=Flag set to true if the attributes object index
			//		used to retrieve the corresponding Attr object has to be
			//		updated (because one Attr object has been removed)
			this->get_device_attr()->remove_attribute(*it, /*update_idx*/true);

			try
			{
				if(it_attr->second.type == Tango::DEV_DOUBLE)
					delete [] (Tango::DevDouble *) it_attr->second.value;
				else if(it_attr->second.type == Tango::DEV_LONG)
					delete [] (Tango::DevLong *) it_attr->second.value;
				else if(it_attr->second.type == Tango::DEV_BOOLEAN)
					delete [] (Tango::DevBoolean *) it_attr->second.value;
				else if(it_attr->second.type == Tango::DEV_STRING)
					delete [] (char *) it_attr->second.value;
				attr_data_map.erase(/**it*/it_attr);
				DEBUG_STREAM << __FUNCTION__ << ": " << *it << " removed from attr_data_map";
			} catch(...)
			{
				WARN_STREAM << __FUNCTION__ << ": ERROR removing " << *it << " from attr_data_map";
			}
			save_configuration();
		}
	}


	/*----- PROTECTED REGION END -----*///	TestDevice::remove
	return;
}

	/*----- PROTECTED REGION ID(TestDevice::namespace_ending) ENABLED START -----*/


	//	Additional Methods
//+------------------------------------------------------------------
/*
 *	method:	TestDevice::create_dynamic_attribute
 * Create and configure an attribute
 */
//+------------------------------------------------------------------
void TestDevice::create_dynamic_attribute(vector<string> config)
{
	if(config.size() == 3 || config.size() == 4)
	{
		Tango::Attr *attr;
		char attrname[256];
		Tango::UserDefaultAttrProp attrprop;

		struct timeval attr_time;
		if(config.size() == 4)
		{
			vector<string> timestamp;
			string_explode(config[3], ".", &timestamp);

			if(timestamp.size() == 2)
			{
				attr_time.tv_sec = atoi(timestamp[0].c_str());
				attr_time.tv_usec = atoi(timestamp[1].c_str());
				if(attr_time.tv_sec == 0)
					gettimeofday(&attr_time,NULL);
			}
			else
			{
				gettimeofday(&attr_time,NULL);
			}
		}
		else
		{
			gettimeofday(&attr_time,NULL);
		}

		strcpy(attrname, config[0].c_str());
		if(attr_data_map.find(attrname) != attr_data_map.end())
			Tango::Except::throw_exception((const char*) "Exception creating attribute",
							string("Attribute '") + string(attrname) + string("' already existing"), __FUNCTION__, Tango::ERR);

		vector<string> values;
		string_explode(config[2], ",", &values);

		map_data_t tmp;
		tmp.size1 = (values.size() > MAX_SPECTRUM_SIZE) ? MAX_SPECTRUM_SIZE : values.size();
		tmp.size2 = 0;
		tmp.max1 = 1;
		tmp.max2 = 1;
		tmp.ts = attr_time;
		size_t value_size = (tmp.size1 > 1) ? MAX_SPECTRUM_SIZE : 1;
		if(config[1] == "bool")
		{
			tmp.type = Tango::DEV_BOOLEAN;
			Tango::DevBoolean *value = new Tango::DevBoolean[value_size];
			size_t ind=0;
			for(vector<string>::iterator val=values.begin(); val != values.end() && ind < tmp.size1; val++)
			{
				if(*val == "true")
					value[ind++] = true;
				else if(*val == "false")
					value[ind++] = false;
				else
					value[ind++] = atoi(val->c_str());
			}
			tmp.value = value;
		}
		else if(config[1] == "long")
		{
			tmp.type = Tango::DEV_LONG;
			Tango::DevLong *value = new Tango::DevLong[value_size];
			size_t ind=0;
			for(vector<string>::iterator val=values.begin(); val != values.end() && ind < tmp.size1; val++)
			{
				value[ind++] = atoi(val->c_str());
			}
			tmp.value = value;
		}
		else if(config[1] == "double")
		{
			tmp.type = Tango::DEV_DOUBLE;
			Tango::DevDouble *value = new Tango::DevDouble[value_size];
			size_t ind=0;
			for(vector<string>::iterator val=values.begin(); val != values.end() && ind < tmp.size1; val++)
			{
				value[ind++] = atof(val->c_str());
			}
			tmp.value = value;
		}
		else if(config[1] == "string")
		{
			tmp.size1 = 1;
			tmp.type = Tango::DEV_STRING;
			char *value = new char[1024];
			memset(value, 0, 1024);
			strncpy(value, config[2].c_str(), 1023);
			tmp.value = value;
		}

		if(tmp.size1 == 1)
			attr = new ScalarDynAttrib(attrname, tmp.type);
		else
			attr = new SpectrumDynAttrib(attrname, tmp.type, MAX_SPECTRUM_SIZE);
		/*sprintf(strtmp,"Dynamic Attribute %s reading channel %d ",attrname,num_ch);
		 if(unit != "")
		 {
		 attrprop.set_unit(unit.c_str());
		 attrprop.set_standard_unit("1.0");
		 attrprop.set_display_unit(unit.c_str());
		 }
		 attrprop.set_format("%.9g");
		 attrprop.set_description(strtmp);*/
		attr->set_default_properties(attrprop);
		add_attribute(attr);
		set_change_event(attrname, true, false);
		set_archive_event(attrname, true, false);
		attr_data_map.insert(make_pair(attrname, tmp));

		try
		{
			if(config[1] == "bool")
			{
				set_write_value(((Tango::DevBoolean *)tmp.value), attrname, tmp.size1);
			}
			else if(config[1] == "long")
			{
				set_write_value(((Tango::DevLong *)tmp.value), attrname, tmp.size1);
			}
			else if(config[1] == "double")
			{
				set_write_value(((Tango::DevDouble *)tmp.value), attrname, tmp.size1);
			}
			else if(config[1] == "string")
			{
				Tango::DevString str_ptr = (char *)tmp.value;
				set_write_value((Tango::DevString)str_ptr, attrname);
			}
		} catch(Tango::DevFailed &e)
		{
			INFO_STREAM << __FUNCTION__ << " error seting write value=" << e.errors[0].desc;
		}
	}
}


//+------------------------------------------------------------------
/*
 *	method:	TestDevice::string_explode
 * Explode a string
 */
//+------------------------------------------------------------------
void TestDevice::string_explode(string str, string separator, vector<string>* results)
{
	size_t found;

	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			results->push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results->push_back(str);
	}

}

//+------------------------------------------------------------------
/*
 *	method:	TestDevice::save_configuration
 * Save configuration on DB
 */
//+------------------------------------------------------------------
void TestDevice::save_configuration()
{
	Tango::Database *db = new Tango::Database();
	try
	{
		db->set_timeout_millis(10000);	//10 s timeout
		vector<string> configuration;
		char conf_str[256];

		map<std::string,map_data_t>::iterator it;
		for(it = attr_data_map.begin(); it != attr_data_map.end(); it++)
		{
			double ts = (double)(it->second.ts.tv_sec) + ((double)it->second.ts.tv_usec)/1000000.0;
			if(it->second.size1 == 1)
			{
				if(it->second.type == Tango::DEV_DOUBLE)
					sprintf(conf_str, "%s:double:%.16e:%f", it->first.c_str(), *((Tango::DevDouble *) it->second.value), ts);
				else if(it->second.type == Tango::DEV_LONG)
					sprintf(conf_str, "%s:long:%d:%f", it->first.c_str(), *((Tango::DevLong *) it->second.value), ts);
				else if(it->second.type == Tango::DEV_BOOLEAN)
					sprintf(conf_str, "%s:bool:%d:%f", it->first.c_str(), *((Tango::DevBoolean *) it->second.value), ts);
				else if(it->second.type == Tango::DEV_STRING)
					sprintf(conf_str, "%s:string:%s:%f", it->first.c_str(), ((char *) it->second.value), ts);
			}
			else if(it->second.size1 > 1)
			{
				stringstream val;
				if(it->second.type == Tango::DEV_DOUBLE)
				{
					for(size_t i=0; i<it->second.size1; i++)
					{
						val << std::setprecision(std::numeric_limits<double>::digits10+2) << ((Tango::DevDouble *) it->second.value)[i];
						if(i < it->second.size1-1)
							val << ",";
					}
					sprintf(conf_str, "%s:double:%s:%f", it->first.c_str(), val.str().c_str(), ts);
				}
				else if(it->second.type == Tango::DEV_LONG)
				{
					for(size_t i=0; i<it->second.size1; i++)
					{
						val << std::setprecision(std::numeric_limits<Tango::DevLong>::digits10+2) << ((Tango::DevLong *) it->second.value)[i];
						if(i < it->second.size1-1)
							val << ",";
					}
					sprintf(conf_str, "%s:long:%s:%f", it->first.c_str(), val.str().c_str(), ts);
				}
				else if(it->second.type == Tango::DEV_BOOLEAN)
				{
					for(size_t i=0; i<it->second.size1; i++)
					{
						val << std::setprecision(std::numeric_limits<Tango::DevBoolean>::digits10+2) << ((Tango::DevBoolean *) it->second.value)[i];
						if(i < it->second.size1-1)
							val << ",";
					}
					sprintf(conf_str, "%s:bool:%s:%f", it->first.c_str(), val.str().c_str(), ts);
				}
				else if(it->second.type == Tango::DEV_STRING)
				{
					//NOT SUPPORTED
				}
			}

			configuration.push_back(conf_str);
			DEBUG_STREAM << __FUNCTION__ << " saving " << conf_str;
		}
		Tango::DbDatum config("Attr_config");
		Tango::DbData db_data_put;
		config << configuration;
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
		//get_db_device()->put_property(db_data_put);
		if(get_state()!=Tango::ON)
			set_state(Tango::ON);
	} catch(Tango::DevFailed &e)
	{
		INFO_STREAM << __FUNCTION__ << " error saving properties=" << e.errors[0].desc;
		set_state(Tango::FAULT);
	}
	delete db;
}


long TestDevice::create_dynamic_command(const char* cmd_name, Tango::CmdArgType type_in, Tango::CmdArgType type_out, long size_in, long size_out)
{

	std::vector< Tango::Command* >& command_list = get_device_class()->get_command_list();
	command_list.push_back(new CmdClass(cmd_name,
			type_in, type_out,
			"LV argin",
			"LV argout",
			Tango::OPERATOR,
			size_in,
			this));

	return 0;
}

template <typename Attr_type> void TestDevice::set_write_value(Attr_type *value, string attrname, long x)
{
	Tango::WAttribute& wattr = get_device_attr()->get_w_attr_by_name(attrname.c_str());
	if(x == 1)
		wattr.set_write_value(*value);
	else
		wattr.set_write_value(value, x);
}

void TestDevice::set_write_value(Tango::DevString value, string attrname)
{
	Tango::WAttribute& wattr = get_device_attr()->get_w_attr_by_name(attrname.c_str());
	wattr.set_write_value(value);
}

/***************************************************************************

  CmdClass class

***************************************************************************/
CORBA::Any *CmdClass::execute(Tango::DeviceImpl *device,const CORBA::Any &in_any)
{

	TANGO_LOG << "CmdClass::" << get_name() << ": entering..." << endl;

	//const Tango::DevVarLongArray	*argin;
	//extract(in_any, argin);

	return new CORBA::Any();

}

/*bool TestDevice::is_LVCmd_allowed(const CORBA::Any &any)
{
	//	End of Generated Code

	//	Re-Start of Generated Code
return true;
}*/

	/*----- PROTECTED REGION END -----*/	//	TestDevice::namespace_ending
} //	namespace
